import path from "node:path";
import { fs, pico, recursiveReaddir, filesize } from "@mango-scripts/utils";
import { gzipSizeSync } from "gzip-size";
import stripAnsi from "./stripAnsi.mjs";
const canReadAsset = (asset) => /\.(js|css)$/.test(asset.name) && !/service-worker\.js/.test(asset.name) && !/precache-manifest\.[0-9a-f]+\.js/.test(asset.name);
const printFileSizesAfterBuild = async (webpackStats, previousSizeMap, buildFolder, maxBundleGzipSize, maxChunkGzipSize) => {
  const root = previousSizeMap.root;
  const sizes = previousSizeMap.sizes;
  const assets = (webpackStats.stats || [webpackStats]).map(
    (stats) => stats.toJson({ all: false, assets: true }).assets.filter((asset) => canReadAsset(asset)).map((asset) => {
      const fileContents = fs.readFileSync(path.join(root, asset.name));
      const size = gzipSizeSync(fileContents);
      const previousSize = sizes[removeFileNameHash(root, asset.name)];
      const difference = getDifferenceLabel(size, previousSize);
      return {
        folder: path.join(
          path.basename(buildFolder),
          path.dirname(asset.name)
        ),
        name: path.basename(asset.name),
        size,
        sizeLabel: filesize(size) + (difference ? " (" + difference + ")" : "")
      };
    })
  ).reduce((single, all) => all.concat(single), []);
  assets.sort((a, b) => b.size - a.size);
  const longestSizeLabelLength = Math.max(
    ...assets.map((a) => stripAnsi(a.sizeLabel).length)
  );
  let suggestBundleSplitting = false;
  for (const asset of assets) {
    let sizeLabel = asset.sizeLabel;
    const sizeLength = stripAnsi(sizeLabel).length;
    if (sizeLength < longestSizeLabelLength) {
      const rightPadding = " ".repeat(longestSizeLabelLength - sizeLength);
      sizeLabel += rightPadding;
    }
    const isMainBundle = asset.name.indexOf("main.") === 0;
    const maxRecommendedSize = isMainBundle ? maxBundleGzipSize : maxChunkGzipSize;
    const isLarge = maxRecommendedSize && asset.size > maxRecommendedSize;
    if (isLarge && path.extname(asset.name) === ".js") {
      suggestBundleSplitting = true;
    }
    console.log(
      "  " + (isLarge ? pico.yellow(sizeLabel) : sizeLabel) + "  " + pico.dim(asset.folder + path.sep) + pico.cyan(asset.name)
    );
  }
  if (suggestBundleSplitting) {
    console.log();
    console.log(
      pico.yellow("The bundle size is significantly larger than recommended.")
    );
    console.log(
      pico.yellow(
        "Consider reducing it with code splitting: https://goo.gl/9VhYWB"
      )
    );
    console.log(
      pico.yellow(
        "You can also analyze the project dependencies: https://goo.gl/LeUzfb"
      )
    );
  }
};
const removeFileNameHash = (buildFolder, fileName) => fileName.replace(buildFolder, "").replace(/\\/g, "/").replace(
  /\/?(.*)(\.[0-9a-f]+)(\.chunk)?(\.js|\.css)/,
  (match, p1, p2, p3, p4) => p1 + p4
);
const getDifferenceLabel = (currentSize, previousSize) => {
  const FIFTY_KILOBYTES = 1024 * 50;
  const difference = currentSize - previousSize;
  const fileSize = !Number.isNaN(difference) ? filesize(difference) : 0;
  if (difference >= FIFTY_KILOBYTES) {
    return pico.red("+" + fileSize);
  } else if (difference < FIFTY_KILOBYTES && difference > 0) {
    return pico.yellow("+" + fileSize);
  } else if (difference < 0) {
    return pico.green(fileSize);
  } else {
    return "";
  }
};
const measureFileSizesBeforeBuild = async (buildFolder) => {
  const fileNames = await recursiveReaddir(buildFolder);
  let sizes = {};
  if (fileNames) {
    sizes = fileNames.filter((i) => canReadAsset(i)).reduce((memo, fileName) => {
      const contents = fs.readFileSync(fileName);
      const key = removeFileNameHash(buildFolder, fileName);
      memo[key] = gzipSizeSync(contents);
      return memo;
    }, {});
  }
  return {
    root: buildFolder,
    sizes: sizes || {}
  };
};
export {
  measureFileSizesBeforeBuild,
  printFileSizesAfterBuild
};
