import path from "node:path";
import os from "node:os";
import { pico } from "@mango-scripts/utils";
class ModuleScopePlugin {
  appSrcs;
  allowedFiles;
  allowedPaths;
  constructor(appSrc, allowedFiles = []) {
    this.appSrcs = Array.isArray(appSrc) ? appSrc : [appSrc];
    this.allowedFiles = new Set(allowedFiles);
    this.allowedPaths = allowedFiles.map(path.dirname).filter((p) => path.relative(p, process.cwd()) !== "");
  }
  apply(resolver) {
    const { appSrcs } = this;
    resolver.hooks.file.tapAsync(
      "ModuleScopePlugin",
      (request, contextResolver, callback) => {
        if (!request.context.issuer) {
          return callback();
        }
        if (
          // If this resolves to a node_module, we don't care what happens next
          request.descriptionFileRoot.indexOf("/node_modules/") !== -1 || request.descriptionFileRoot.indexOf("\\node_modules\\") !== -1 || // Make sure this request was manual
          !request.__innerRequest_request
        ) {
          return callback();
        }
        if (appSrcs.every((appSrc) => {
          const relative = path.relative(appSrc, request.context.issuer);
          return relative.startsWith("../") || relative.startsWith("..\\");
        })) {
          return callback();
        }
        const requestFullPath = path.resolve(
          path.dirname(request.context.issuer),
          request.__innerRequest_request
        );
        if (this.allowedFiles.has(requestFullPath)) {
          return callback();
        }
        if (this.allowedPaths.some((allowedFile) => {
          return requestFullPath.startsWith(allowedFile);
        })) {
          return callback();
        }
        if (appSrcs.every((appSrc) => {
          const requestRelative = path.relative(appSrc, requestFullPath);
          return requestRelative.startsWith("../") || requestRelative.startsWith("..\\");
        })) {
          const scopeError = new Error(
            `You attempted to import ${pico.cyan(
              request.__innerRequest_request
            )} which falls outside of the project ${pico.cyan(
              "src/"
            )} directory. Relative imports outside of ${pico.cyan(
              "src/"
            )} are not supported.` + os.EOL + `You can either move it inside ${pico.cyan(
              "src/"
            )}, or add a symlink to it from project's ${pico.cyan(
              "node_modules/"
            )}.`
          );
          Object.defineProperty(scopeError, "__module_scope_plugin", {
            value: true,
            writable: false,
            enumerable: false
          });
          callback(scopeError, request);
        } else {
          callback();
        }
      }
    );
  }
}
var ModuleScopePlugin_default = ModuleScopePlugin;
export {
  ModuleScopePlugin_default as default
};
