import path from "node:path";
import url from "node:url";
import { fs, pico } from "@mango-scripts/utils";
import { ip } from "address";
import ForkTsCheckerWebpackPlugin from "fork-ts-checker-webpack-plugin";
import formatWebpackMessages from "./formatWebpackMessages.mjs";
import clearConsole from "./clearConsole.mjs";
const isInteractive = process.stdout.isTTY;
const prepareUrls = (protocol, host, port, pathname = "/") => {
  const formatUrl = (hostname) => url.format({
    protocol,
    hostname,
    port,
    pathname
  });
  const prettyPrintUrl = (hostname) => url.format({
    protocol,
    hostname,
    port: pico.bold(port.toString()),
    pathname
  });
  const isUnspecifiedHost = host === "0.0.0.0" || host === "::";
  let prettyHost, lanUrlForConfig, lanUrlForTerminal;
  if (isUnspecifiedHost) {
    prettyHost = "localhost";
    try {
      lanUrlForConfig = ip();
      if (lanUrlForConfig) {
        if (/^10[.]|^172[.](1[6-9]|2[0-9]|3[0-1])[.]|^192[.]168[.]/.test(
          lanUrlForConfig
        )) {
          lanUrlForTerminal = prettyPrintUrl(lanUrlForConfig);
        } else {
          lanUrlForConfig = void 0;
        }
      }
    } catch (_e) {
    }
  } else {
    prettyHost = host;
  }
  const localUrlForTerminal = prettyPrintUrl(prettyHost);
  const localUrlForBrowser = formatUrl(prettyHost);
  return {
    lanUrlForConfig,
    lanUrlForTerminal,
    localUrlForTerminal,
    localUrlForBrowser
  };
};
const printInstructions = (appName, urls) => {
  console.log(`
You can now view ${pico.bold(appName)} in the browser.
`);
  if (urls.lanUrlForTerminal) {
    console.log(
      `  ${pico.bold("Local:")}            ${urls.localUrlForTerminal}`
    );
    console.log(`  ${pico.bold("On Your Network:")}  ${urls.lanUrlForTerminal}`);
  } else {
    console.log(`  ${urls.localUrlForTerminal}`);
  }
  console.log("\nNote that the development build is not optimized.");
  console.log(
    `To create a production build, use ${pico.cyan(`pnpm run build`)}.
`
  );
};
const createCompiler = ({
  appName,
  config,
  urls,
  webpack
}) => {
  let compiler;
  try {
    compiler = webpack(config);
  } catch (err) {
    console.log(pico.red("Failed to compile.\n"));
    console.log(err.message || err);
    console.log();
    process.exit(1);
  }
  compiler.hooks.invalid.tap("invalid", () => {
    if (isInteractive) {
      clearConsole();
    }
    console.log("Compiling...");
  });
  let isFirstCompile = true;
  let tsMessagesPromise;
  ForkTsCheckerWebpackPlugin.getCompilerHooks(compiler).waiting.tap(
    "awaitingTypeScriptCheck",
    () => {
      console.log(
        pico.yellow(
          "Files successfully emitted, waiting for typecheck results..."
        )
      );
    }
  );
  compiler.hooks.done.tap("done", async (stats) => {
    if (isInteractive) {
      clearConsole();
    }
    const statsData = stats.toJson({
      all: false,
      warnings: true,
      errors: true
    });
    const messages = formatWebpackMessages(statsData);
    const isSuccessful = !messages.errors.length && !messages.warnings.length;
    if (isSuccessful) {
      console.log(pico.green("Compiled successfully!"));
    }
    if (isSuccessful && (isInteractive || isFirstCompile)) {
      printInstructions(appName, urls);
    }
    isFirstCompile = false;
    if (messages.errors.length) {
      if (messages.errors.length > 1) {
        messages.errors.length = 1;
      }
      console.log(pico.red("Failed to compile.\n"));
      console.log(messages.errors.join("\n\n"));
      return;
    }
    if (messages.warnings.length) {
      console.log(pico.yellow("Compiled with warnings.\n"));
      console.log(messages.warnings.join("\n\n"));
      console.log(
        "\nSearch for the " + pico.underline(pico.yellow("keywords")) + " to learn more about each warning."
      );
      console.log(
        "To ignore, add " + pico.cyan("// eslint-disable-next-line") + " to the line before.\n"
      );
    }
  });
  const isSmokeTest = process.argv.some(
    (arg) => arg.indexOf("--smoke-test") > -1
  );
  if (isSmokeTest) {
    compiler.hooks.failed.tap("smokeTest", async () => {
      await tsMessagesPromise;
      process.exit(1);
    });
    compiler.hooks.done.tap("smokeTest", async (stats) => {
      await tsMessagesPromise;
      if (stats.hasErrors() || stats.hasWarnings()) {
        process.exit(1);
      } else {
        process.exit(0);
      }
    });
  }
  return compiler;
};
const resolveLoopback = (proxy) => {
  const o = url.parse(proxy);
  o.host = void 0;
  if (o.hostname !== "localhost") {
    return proxy;
  }
  try {
    if (!ip()) {
      o.hostname = "127.0.0.1";
    }
  } catch (_ignored) {
    o.hostname = "127.0.0.1";
  }
  return url.format(o);
};
const onProxyError = (proxy) => (err, req, res) => {
  const host = req.headers && req.headers.host;
  console.log(
    pico.red("Proxy error:") + " Could not proxy request " + pico.cyan(req.url) + " from " + pico.cyan(host) + " to " + pico.cyan(proxy) + "."
  );
  console.log(
    "See https://nodejs.org/api/errors.html#errors_common_system_errors for more information (" + pico.cyan(err.code) + ")."
  );
  console.log();
  if (res.writeHead && !res.headersSent) {
    res.writeHead(500);
  }
  res.end(
    "Proxy error: Could not proxy request " + req.url + " from " + host + " to " + proxy + " (" + err.code + ")."
  );
};
const prepareProxy = (proxy, appPublicFolder, servedPathname) => {
  if (!proxy) {
    return void 0;
  }
  if (typeof proxy !== "string") {
    console.log(
      pico.red('When specified, "proxy" in package.json must be a string.')
    );
    console.log(
      pico.red('Instead, the type of "proxy" was "' + typeof proxy + '".')
    );
    console.log(
      pico.red('Either remove "proxy" from package.json, or make it a string.')
    );
    process.exit(1);
  }
  const sockPath = process.env.WDS_SOCKET_PATH || "/ws";
  const isDefaultSockHost = !process.env.WDS_SOCKET_HOST;
  const mayProxy = (pathname) => {
    const maybePublicPath = path.resolve(
      appPublicFolder,
      pathname.replace(new RegExp("^" + servedPathname), "")
    );
    const isPublicFileRequest = fs.existsSync(maybePublicPath);
    const isWdsEndpointRequest = isDefaultSockHost && pathname.startsWith(sockPath);
    return !(isPublicFileRequest || isWdsEndpointRequest);
  };
  if (!/^http(s)?:\/\//.test(proxy)) {
    console.log(
      pico.red(
        'When "proxy" is specified in package.json it must start with either http:// or https://'
      )
    );
    process.exit(1);
  }
  let target;
  if (process.platform === "win32") {
    target = resolveLoopback(proxy);
  } else {
    target = proxy;
  }
  return [
    {
      target,
      logLevel: "silent",
      // For single page apps, we generally want to fallback to /index.html.
      // However we also want to respect `proxy` for API calls.
      // So if `proxy` is specified as a string, we need to decide which fallback to use.
      // We use a heuristic: We want to proxy all the requests that are not meant
      // for static assets and as all the requests for static assets will be using
      // `GET` method, we can proxy all non-`GET` requests.
      // For `GET` requests, if request `accept`s text/html, we pick /index.html.
      // Modern browsers include text/html into `accept` header when navigating.
      // However API calls like `fetch()` won’t generally accept text/html.
      // If this heuristic doesn’t work well for you, use `src/setupProxy.js`.
      context: (pathname, req) => {
        return req.method !== "GET" || mayProxy(pathname) && req.headers.accept && req.headers.accept.indexOf("text/html") === -1;
      },
      onProxyReq: (proxyReq) => {
        if (proxyReq.getHeader("origin")) {
          proxyReq.setHeader("origin", target);
        }
      },
      onError: onProxyError(target),
      secure: false,
      changeOrigin: true,
      ws: true,
      xfwd: true
    }
  ];
};
export {
  createCompiler,
  prepareProxy,
  prepareUrls
};
